@@ -1,4 +1,4 @@
-### Copyright (C) 2002-2003 Stephen Kennedy <stevek@gnome.org>
+### Copyright (C) 2002-2004 Stephen Kennedy <stevek@gnome.org>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -35,6 +35,45 @@ import misc
 import melddoc
 import paths
 
+sourceview_available = 0
+
+for sourceview in "gtksourceview sourceview".split():
+    try:
+        gsv = __import__(sourceview)
+        sourceview_available = 1
+        break
+    except ImportError:
+        pass
+
+if sourceview_available:
+    def set_highlighting_enabled(buf, fname, enabled):
+        # gnome.vfs.get_mime_type seems to be broken. fake it.
+        extmap = { "xml":"text/xml",
+                   "glade":"text/xml",
+                   "cpp":"text/x-cpp",
+                   "cxx":"text/x-cpp",
+                   "cc":"text/x-cpp",
+                   "C":"text/x-cpp",
+                   "c":"text/x-c",
+                   "hpp":"text/x-cpp",
+                   "hxx":"text/x-cpp",
+                   "hh":"text/x-cpp",
+                   "H":"text/x-cpp",
+                   "h":"text/x-cpp",
+                   "inl":"text/x-cpp",
+                   "desktop": "application/x-desktop",
+                   "diff": "text/x-diff",
+                   "patch": "text/x-diff",
+                   "html": "text/html",
+                   "po": "text/x-po",
+                   "py": "text/x-python" }
+        ext = fname.split(".")[-1]
+        man = gsv.SourceLanguagesManager()
+        gsl = man.get_language_from_mime_type( extmap.get(ext, "text/plain") )
+        if gsl:
+            buf.set_language(gsl)
+        buf.set_highlight(enabled)
+
 gdk = gtk.gdk
 
 ################################################################################
@@ -51,7 +90,10 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
 
     keylookup = {gtk.keysyms.Shift_L : MASK_SHIFT,
                  gtk.keysyms.Control_L : MASK_CTRL,
-                 gtk.keysyms.Alt_L : MASK_ALT}
+                 gtk.keysyms.Alt_L : MASK_ALT,
+                 gtk.keysyms.Shift_R : MASK_SHIFT,
+                 gtk.keysyms.Control_R : MASK_CTRL,
+                 gtk.keysyms.Alt_R : MASK_ALT }
 
     def __init__(self, prefs, num_panes):
         """Start up an filediff with num_panes empty contents.
@@ -59,9 +101,30 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         melddoc.MeldDoc.__init__(self, prefs)
         gnomeglade.Component.__init__(self, paths.share_dir("glade2/filediff.glade"), "filediff")
         self._map_widgets_into_lists( ["textview", "fileentry", "diffmap", "scrolledwindow", "linkmap", "statusimage"] )
+        self._update_regexes()
+        self.warned_bad_comparison = False
+        if sourceview_available:
+            # ugly hack. http://bugzilla.gnome.org/show_bug.cgi?id=140071
+            # will remove the need for this.
+            self.textview = []
+            for w in self.scrolledwindow:
+                w.remove( w.get_child() )
+                v = gsv.SourceView()
+                self.textview.append( v )
+                v.show()
+                w.add(v)
+                v.set_show_line_numbers(self.prefs.show_line_numbers)
+                for s in "key_press_event key_release_event".split():
+                    v.connect(s, getattr(self,"on_%s"%s))
+                for s in "button_press_event focus_in_event".split():
+                    v.connect(s, getattr(self,"on_textview_%s"%s))
+                v.get_buffer().connect("mark-set", self.on_textbuffer_mark_set)
         self.keymask = 0
         self.load_font()
         self.deleted_lines_pending = -1
+        self.textview_overwrite = 0
+        self.textview_focussed = None
+        self.textview_overwrite_handlers = [ t.connect("toggle-overwrite", self.on_textview_toggle_overwrite) for t in self.textview ]
         for i in range(3):
             w = self.scrolledwindow[i]
             w.get_vadjustment().connect("value-changed", self._sync_vscroll )
@@ -70,8 +133,10 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.linediffer = diffutil.Differ()
         for l in self.linkmap: # glade bug workaround
             l.set_events(gdk.BUTTON_PRESS_MASK | gdk.BUTTON_RELEASE_MASK )
+            l.set_double_buffered(0) # we call paint_begin ourselves
         self.bufferdata = []
         for text in self.textview:
+            text.set_wrap_mode( self.prefs.edit_wrap_lines )
             buf = text.get_buffer()
             self.bufferdata.append( MeldBufferData() )
             def add_tag(name, props):
@@ -121,6 +186,15 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.last_search = None
         self.set_num_panes(num_panes)
 
+    def _update_regexes(self):
+        self.regexes = []
+        for r in [ misc.ListItem(i) for i in self.prefs.regexes.split("\n") ]:
+            if r.active:
+                try:
+                    self.regexes.append( (re.compile(r.value+"(?m)"), r.value) )
+                except re.error, e:
+                    pass
+
     def _disconnect_buffer_handlers(self):
         for textview in self.textview:
             buf = textview.get_buffer()
@@ -140,29 +214,71 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             buf.textview = textview
             buf.handlers = id0, id1, id2, id3
 
+    def _update_cursor_status(self, buf):
+        def update():
+            iter = buf.get_iter_at_mark( buf.get_insert() )
+            status = "%s : %s" % ( _("INS,OVR").split(",")[ self.textview_overwrite ], #insert/overwrite
+                                   _("Ln %i, Col %i") % (iter.get_line()+1, iter.get_line_offset()+1) ) #line/column
+            self.emit("status-changed", status  )
+            raise StopIteration; yield 0
+        self.scheduler.add_task( update().next )
+
+    def on_textbuffer_mark_set(self, buffer, it, mark):
+        if mark.get_name() == "insert":
+            self._update_cursor_status(buffer)
+    def on_textview_focus_in_event(self, view, event):
+        self.textview_focussed = view
+        self._update_cursor_status(view.get_buffer())
+    def on_switch_event(self):
+        if self.textview_focussed:
+            self.scheduler.add_task( self.textview_focussed.grab_focus )
+
     def _after_text_modified(self, buffer, startline, sizechange):
         if self.num_panes > 1:
             buffers = [t.get_buffer() for t in self.textview[:self.num_panes] ]
             pane = buffers.index(buffer)
             range = self.linediffer.change_sequence( pane, startline, sizechange, self._get_texts())
-            #print "***", range, len(self.linediffer.diffs[0])
             for iter in self._update_highlighting( range[0], range[1] ):
                 pass
             self.queue_draw()
+        self._update_cursor_status(buffer)
 
-    def _get_texts(self):
+    def _get_texts(self, raw=0):
         class FakeText(object):
-            def __init__(self, buf):
-                self.buf = buf
+            def __init__(self, buf, textfilter):
+                self.buf, self.textfilter = buf, textfilter
             def __getslice__(self, lo, hi):
                 b = self.buf
-                return b.get_text(b.get_iter_at_line(lo), b.get_iter_at_line(hi), 0).split("\n")[:-1]
+                txt = b.get_text(b.get_iter_at_line(lo), b.get_iter_at_line(hi), 0)
+                txt = self.textfilter(txt)
+                return txt.split("\n")[:-1]
         class FakeTextArray(object):
-            def __init__(self, bufs):
-                self.texts = map(FakeText, bufs)
+            def __init__(self, bufs, textfilter):
+                self.texts = [FakeText(b, textfilter) for b in  bufs]
             def __getitem__(self, i):
                 return self.texts[i]
-        return FakeTextArray( [t.get_buffer() for t in self.textview] )
+        return FakeTextArray( [t.get_buffer() for t in self.textview], [self._filter_text, lambda x:x][raw] )
+
+    def _filter_text(self, txt):
+        def killit(m):
+            assert m.group().count("\n") == 0
+            if len(m.groups()):
+                s = m.group()
+                for g in m.groups():
+                    if g:
+                        s = s.replace(g,"")
+                return s
+            else:
+                return ""
+        try:
+            for c,r in self.regexes:
+                txt = c.sub(killit,txt)
+        except AssertionError:
+            if self.warned_bad_comparison == False:
+                misc.run_dialog(_("Regular expression '%s' changed the number of lines in the file. " \
+                    "Comparison will be incorrect. See the user manual for more details.") % r)
+                self.warned_bad_comparison = True
+        return txt
 
     def after_text_insert_text(self, buffer, iter, newtext, textlen):
         lines_added = newtext.count("\n")
@@ -209,6 +325,19 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 self.textview[i].set_tabs(tabs)
         elif key == "use_custom_font" or key == "custom_font":
             self.load_font()
+        elif key == "show_line_numbers":
+            if sourceview_available:
+                for t in self.textview:
+                    t.set_show_line_numbers( value )
+        elif key == "use_syntax_highlighting":
+            if sourceview_available:
+                for i in range(self.num_panes):
+                    set_highlighting_enabled(
+                        self.textview[i].get_buffer(),
+                        self.bufferdata[i].filename,
+                        self.prefs.use_syntax_highlighting )
+        elif key == "regexes":
+            self._update_regexes()
 
     def on_key_press_event(self, object, event):
         x = self.keylookup.get(event.keyval, 0)
@@ -245,6 +374,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             buttons = []
             for i in range(self.num_panes):
                 b = gtk.CheckButton( self._get_filename(i) )
+                b.set_use_underline(False)
                 buttons.append(b)
                 dialog.box.pack_start(b, 1, 1)
                 if not modified[i]:
@@ -310,6 +440,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         return None
 
     def on_find_activate(self, *args):
+        self.keymask = 0
+        self.queue_draw()
         if self.find_dialog:
             self.find_dialog.widget.present()
         else:
@@ -362,10 +494,14 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.popup_menu.popup_in_pane( pane )
             return 1
 
-    #def on_textview_toggle_overwrite(self, view):
-        #print view.overwrite_mode
-        #print misc.ilook("over", view)
-        #for v in view:
+    def on_textview_toggle_overwrite(self, view):
+        self.textview_overwrite = not self.textview_overwrite
+        for v,h in zip(self.textview, self.textview_overwrite_handlers):
+            v.disconnect(h)
+            if v != view:
+                v.emit("toggle-overwrite")
+        self.textview_overwrite_handlers = [ t.connect("toggle-overwrite", self.on_textview_toggle_overwrite) for t in self.textview ]
+        self._update_cursor_status(view.get_buffer())
 
         #
         # find/replace buffer
@@ -505,6 +641,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             yield 1
         self.undosequence.clear()
         yield _("[%s] Computing differences") % self.label_text
+        panetext = [self._filter_text(p) for p in panetext]
         lines = map(lambda x: x.split("\n"), panetext)
         step = self.linediffer.set_sequences_iter(*lines)
         while step.next() == None:
@@ -513,34 +650,13 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         lenseq = [len(d) for d in self.linediffer.diffs]
         self.scheduler.add_task( self._update_highlighting( (0,lenseq[0]), (0,lenseq[1]) ).next )
         self._connect_buffer_handlers()
+        if sourceview_available:
+            for i in range(len(files)):
+                if files[i]:
+                    set_highlighting_enabled( self.textview[i].get_buffer(), files[i], self.prefs.use_syntax_highlighting )
         yield 0
 
     def _update_highlighting(self, range0, range1):
-        if 0:
-            def line_range(diffs, range, offset, linerange):
-                if len(diffs):
-                    if range[0] > 0:
-                        chunk = diffs[range[0]-1]
-                        l0c = chunk[offset+1]
-                    else:
-                        l0c = linerange[0]
-                    if range[1] < len(diffs):
-                        chunk = diffs[range0[1]]
-                        h0c = chunk[offset]
-                    else:
-                        h0c = linerange[1]
-                    return l0c, h0c
-                return None
-            ranges = [None, None, None, None]
-            ranges[0] = line_range(self.linediffer.diffs[0], range0, 3, (0, self.linediffer.seqlength[0]) )
-            ranges[1] = line_range(self.linediffer.diffs[0], range0, 1, (0, self.linediffer.seqlength[1]) )
-            ranges[2] = line_range(self.linediffer.diffs[1], range1, 1, (0, self.linediffer.seqlength[1]) )
-            ranges[3] = line_range(self.linediffer.diffs[1], range1, 3, (0, self.linediffer.seqlength[2]) )
-            buffers = [t.get_buffer() for t in self.textview]
-            buffers.insert(1, buffers[1])
-            buffers = [t.get_buffer() for t in self.textview]
-            ranges = [(0, length) for length in self.linediffer.seqlength]
-        #for b, l in zip( buffers, ranges ):
         buffers = [t.get_buffer() for t in self.textview]
         for b in buffers:
             taglist = ["delete line", "conflict line", "replace line", "inline line"]
@@ -548,12 +664,15 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
             for tagname in taglist:
                 tag = table.lookup(tagname)
                 b.remove_tag(tag, b.get_start_iter(), b.get_end_iter() )
-        #for chunk in self.linediffer.all_changes_in_range(self._get_texts(), range0[0], range0[1], range1[0], range1[1]):
         for chunk in self.linediffer.all_changes(self._get_texts()):
             for i,c in misc.enumerate(chunk):
                 if c:
                     if c[0] == "insert":
                         buf = buffers[i*2]
+                        #txt = self._get_texts()[0][c[3]:c[4]]
+                        #print txt, "".join(txt) == ""
+                        #if "".join(txt) == "": continue
+                        #print "OK"
                         tag = buf.get_tag_table().lookup("delete line")
                         buf.apply_tag( tag, buf.get_iter_at_line(c[3]), buf.get_iter_at_line(c[4]) )
                     elif c[0] == "delete":
@@ -572,8 +691,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                         for b, t, s, l in zip(bufs, tags, starts, (c[2],c[4])):
                             b.apply_tag(t, s, b.get_iter_at_line(l))
                         if 1:
-                            text1 = "\n".join( self._get_texts()[1  ][c[1]:c[2]] )
-                            textn = "\n".join( self._get_texts()[i*2][c[3]:c[4]] )
+                            text1 = "\n".join( self._get_texts(raw=1)[1  ][c[1]:c[2]] )
+                            textn = "\n".join( self._get_texts(raw=1)[i*2][c[3]:c[4]] )
                             matcher = difflib.SequenceMatcher(None, text1, textn)
                             #print "<<<\n%s\n---\n%s\n>>>" % (text1, textn)
                             tags = [b.get_tag_table().lookup("inline line") for b in bufs]
@@ -632,14 +751,20 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                         if k != '\n':
                             text = text.replace('\n', k)
                         break
-        if bufdata.encoding:
-            text = text.encode(bufdata.encoding)
+        if bufdata.encoding and self.prefs.save_encoding==0:
+            try:
+                text = text.encode(bufdata.encoding)
+            except UnicodeEncodeError, e:
+                if misc.run_dialog(
+                    _("'%s' contains characters not encodable with '%s'\nWould you like to save as UTF-8?") % (bufdata.filename, bufdata.encoding),
+                    self, gtk.MESSAGE_ERROR, gtk.BUTTONS_YES_NO) != gtk.RESPONSE_YES:
+                    return melddoc.RESULT_ERROR
         try:
             open(bufdata.filename, "w").write(text)
         except IOError, e:
             misc.run_dialog(
                 _("Error writing to %s\n\n%s.") % (bufdata.filename, e),
-                self, buttongtk.MESSAGE_ERROR, gtk.BUTTONS_OK)
+                self, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK)
             return melddoc.RESULT_ERROR
         else:
             self.emit("file-changed", bufdata.filename)
@@ -692,7 +817,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         self.scheduler.add_task( lambda : self._sync_vscroll( self.scrolledwindow[src_pane].get_vadjustment() ) and None )
 
         #
-        # refresh, _queue_refresh
+        # refresh
         #
     def refresh(self, junk=None):
         modified = [b.filename for b in self.bufferdata if b.modified]
@@ -858,7 +983,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         if direction == gdk.SCROLL_DOWN:
             for c in self.linediffer.single_changes(1, self._get_texts()):
                 assert c[0] != "equal"
-                if c[1] > curline:
+                if c[1] > curline + 1:
                     break
         else: #direction == gdk.SCROLL_UP
             for chunk in self.linediffer.single_changes(1, self._get_texts()):
@@ -892,11 +1017,19 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         area.meldgc = misc.struct(gc_delete=gcd, gc_insert=gcd, gc_replace=gcc, gc_conflict=gcx)
         area.meldgc.get_gc = lambda p: getattr(area.meldgc, "gc_"+p)
 
-    def on_textview_expose_event(self, textview, event):
-        gcd = textview.window.new_gc()
-        points = [(0,0), (100,100)]
-        textview.window.draw_lines(gcd, points  )
-        #print points
+    def _consume_blank_lines(self, txt):
+        lo, hi = 0, 0
+        for l in txt:
+            if len(l)==0:
+                lo += 1
+            else:
+                break
+        for l in txt[lo:]:
+            if len(l)==0:
+                hi += 1
+            else:
+                break
+        return lo,hi
 
         #
         # linkmap drawing
@@ -905,12 +1038,13 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         window = area.window
         # not mapped? 
         if not window: return
-        alloc = area.get_allocation()
-        (wtotal,htotal) = alloc.width, alloc.height
-        
         if not hasattr(area, "meldgc"):
             self._setup_gcs(area)
         gctext = area.get_style().text_gc[0]
+
+        alloc = area.get_allocation()
+        (wtotal,htotal) = alloc.width, alloc.height
+        window.begin_paint_rect( (0,0,wtotal,htotal) )
         window.clear()
 
         # gain function for smoothing
@@ -945,6 +1079,13 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         visible = [None] + bounds(which) + bounds(which+1)
 
         for c in self.linediffer.pair_changes(which, which+1, self._get_texts()):
+            if self.prefs.ignore_blank_lines:
+                c1,c2 = self._consume_blank_lines( self._get_texts()[which  ][c[1]:c[2]] )
+                c3,c4 = self._consume_blank_lines( self._get_texts()[which+1][c[3]:c[4]] )
+                c = c[0], c[1]+c1,c[2]-c2, c[3]+c3,c[4]-c4
+                if c[1]==c[2] and c[3]==c[4]:
+                    continue
+
             assert c[0] != "equal"
             if c[2] < visible[1] and c[4] < visible[3]: # find first visible chunk
                 continue
@@ -967,8 +1108,8 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 points = points0 + points1 + [points0[0]]
 
                 window.draw_polygon( gc(c[0]), 1, points)
-                window.draw_lines(gctext, points0  )
-                window.draw_lines(gctext, points1  )
+                window.draw_lines(gctext, points0)
+                window.draw_lines(gctext, points1)
             else:
                 w = wtotal
                 p = self.pixbuf_apply0.get_width()
@@ -989,6 +1130,7 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
         # allow for scrollbar at end of textview
         mid = 0.5 * self.textview0.get_allocation().height
         window.draw_line(gctext, int(.25*wtotal), int(mid), int(.75*wtotal), int(mid) )
+        window.end_paint()
 
     def on_linkmap_scroll_event(self, area, event):
         self.next_diff(event.direction)
@@ -1020,12 +1162,18 @@ class FileDiff(melddoc.MeldDoc, gnomeglade.Component):
                 rect_x = wtotal - pix_width
             else:
                 return  1
-            adj = self.scrolledwindow[which+side].get_vadjustment()
-            func = lambda c: self._line_to_pixel(which, c[1]) - adj.value
-
             src = which + side
             dst = which + 1 - side
+            adj = self.scrolledwindow[src].get_vadjustment()
+            func = lambda c: self._line_to_pixel(src, c[1]) - adj.value
+
             for c in self.linediffer.pair_changes(src, dst, self._get_texts()):
+                if self.prefs.ignore_blank_lines:
+                    c1,c2 = self._consume_blank_lines( self._get_texts()[src][c[1]:c[2]] )
+                    c3,c4 = self._consume_blank_lines( self._get_texts()[dst][c[3]:c[4]] )
+                    c = c[0], c[1]+c1,c[2]-c2, c[3]+c3,c[4]-c4
+                    if c[1]==c[2] and c[3]==c[4]:
+                        continue
                 if c[0] == "insert":
                     continue
                 h = func(c)
