@@ -1,4 +1,4 @@
-### Copyright (C) 2002-2003 Stephen Kennedy <stevek@gnome.org>
+### Copyright (C) 2002-2004 Stephen Kennedy <stevek@gnome.org>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -51,7 +51,6 @@ def run_dialog( text, parent=None, messagetype=gtk.MESSAGE_WARNING, buttonstype=
         d.set_transient_for(parent.widget.get_toplevel())
     for b,id in extrabuttons:
         d.add_button(b,id)
-    d.set_has_separator(0)
     d.vbox.set_spacing(12)
     hbox = d.vbox.get_children()[0]
     hbox.set_spacing(12)
@@ -139,8 +138,8 @@ def read_pipe_iter(command, errorstream, yield_interval=0.1, workdir=None):
     """Read the output of a shell command iteratively.
 
     Each time 'callback_interval' seconds pass without reading any data,
-    this function yeilds None.
-    When all the data is read, the entire string is yeilded.
+    this function yields None.
+    When all the data is read, the entire string is yielded.
     If 'workdir' is specified the command is run from that directory.
     """
     class sentinel(object):
@@ -238,11 +237,19 @@ def commonprefix(dirs):
     return os.sep.join(prefix)
 
 def escape(s):
-    """Replace special characters '&', '<' and '>' by SGML entities.
+    """Replace special characters by SGML entities.
     """
-    s = s.replace("&", "&amp;")
-    s = s.replace("<", "&lt;")
-    s = s.replace(">", "&gt;")
+    entities = ("&&amp;", "<&lt;", ">&gt;")
+    for e in entities:
+        s = s.replace(e[0], e[1:])
+    return s
+
+def unescape(s):
+    """Inverse of escape.
+    """
+    entities = (">&gt;", "<&lt;", "&&amp;")
+    for e in entities:
+        s = s.replace(e[1:], e[0])
     return s
 
 def copytree(src, dst, symlinks=1):
