@@ -1,4 +1,4 @@
-### Copyright (C) 2002-2003 Stephen Kennedy <stevek@gnome.org>
+### Copyright (C) 2002-2004 Stephen Kennedy <stevek@gnome.org>
 
 ### This program is free software; you can redistribute it and/or modify
 ### it under the terms of the GNU General Public License as published by
@@ -107,8 +107,11 @@ class Preferences(object):
             value.current = val
             setfunc = getattr(self._gconf, "set_%s" % value.type)
             setfunc("%s/%s" % (self._rootkey, attr), val)
-            for l in self._listeners:
-                l(attr,val)
+            try:
+                for l in self._listeners:
+                    l(attr,val)
+            except StopIteration:
+                pass
 
     def _on_preference_changed(self, client, timestamp, entry, extra):
         attr = entry.key[ entry.key.rindex("/")+1 : ]
